unit CDTools;

{-------------------------------------------------------------------------------
  CDTools
  funkce pro praci s jednotkou CD-ROM
  2003 (c) manison
-------------------------------------------------------------------------------}

interface

uses Windows, SysUtils, Classes, MMSystem, ComObj, Forms, ShlObj, FileCtrl;

type
  ECdRom = class(Exception);

{ naplni seznam dostupnymi jednotkami CD-ROM }
procedure ListCdDrives(List: TStrings);

{ vrac informaci o tom, zda se v jednotce nachazi disk (je mozne pouzit i pro
  disketove mechaniky) }
function MediumPresent(const Root: string): Boolean;

{ ziska seriove cislo a jmenovku disku (je mozne pouzit pro jakykoliv disk) }
procedure GetMediumInfo(const Root: string; var Serial: Cardinal;
  var Name: string);

type
  TCdDoorOp = (cdOpen, cdClose);

{ otevre/zavre dvirka }
procedure OpenCloseCd(const Root: string; Operation: TCdDoorOp);

type
  DEV_BROADCAST_HDR = packed record
    dbch_size: DWORD;
    dbch_devicetype: DWORD;
    dbch_reserved: DWORD;
  end;
  TDevBroadcastHdr = DEV_BROADCAST_HDR;
  PDevBroadcastHdr = ^TDevBroadcastHdr;

  DEV_BROADCAST_VOLUME = packed record
    dbcv_size: DWORD;
    dbcv_devicetype: DWORD;
    dbcv_reserved: DWORD;
    dbcv_unitmask: DWORD;
    dbcv_flags: WORD;
  end;
  TDevBroadcastVolume = DEV_BROADCAST_VOLUME;
  PDevBroadcastVolume = ^TDevBroadcastVolume;
  
const
  DBT_DEVICEARRIVAL = $8000;
  DBT_DEVICEQUERYREMOVE = $8001;
  DBT_DEVICEQUERYREMOVEFAILED = $8002;
  DBT_DEVICEREMOVEPENDING = $8003;
  DBT_DEVICEREMOVECOMPLETE = $8004;

  DBT_DEVTYP_VOLUME = $00000002;

  DBTF_MEDIA = $0001;

  BROADCAST_QUERY_DENY = $424D5144;

{ vrati pismeno disku odpovidajici bitove masce }
function DriveLetterFromMask(Mask: Cardinal): Char;

{ Windows XP }

type
  ICDBurn = interface(IUnknown)
  ['{3d73a659-e5d0-4d42-afc0-5121ba425c8d}']
    function GetRecorderDriveLetter(pszDrive: LPWSTR; cch: UINT): HResult; stdcall;
    function Burn(hwnd: HWND): HResult; stdcall;
    function HasRecordableDrive(out pfHasRecorder: BOOL): HResult; stdcall;
  end;

const
  CLSID_CDBurn: TGUID = '{fbeb8a05-beee-4442-804e-409d6c4515e9}';

  CSIDL_CDBURN_AREA = $003b;

{ zjist, zda se v pocitaci nachazi jednotka pro zapis na CD }
function CdRwPresent(var Root: string): Boolean;

{ zahaji zapis na CD }
function CdRwBurn: Boolean;

{ vrati slozku, kam jsou ukladany soubory pred vypalenim na CD }
function CdRwPath: string;

type
  DIOC_REGISTERS = packed record
    EBX, EDX, ECX, EAX, EDI, ESI, Flags: DWORD;
  end;
  TDIOCRegisters = DIOC_REGISTERS;
  PDIOCRegisters = ^TDIOCRegisters;
  PDIOC_REGISTERS = PDIOCRegisters;

  PARAMBLOCK = packed record
    bOperation, bNumLocks: Byte;
  end;
  TParamBlock = PARAMBLOCK;
  PParamBlock = ^TParamBlock;

const
  IOCTL_STORAGE_EJECT_MEDIA =  $002D4808;
  IOCTL_STORAGE_LOAD_MEDIA = $002D480C;
  IOCTL_STORAGE_MEDIA_REMOVAL = $002D4804;
  IOCTL_STORAGE_EJECTION_CONTROL = $002D0940;
  FSCTL_LOCK_VOLUME = $00090018;
  FSCTL_DISMOUNT_VOLUME = $00090020;

  VWIN32_DIOC_DOS_IOCTL = 1;

  FLAG_CARRY = 1;

{ otevre / zavre dvirka pres IO control kody }
function OpenCloseCdIoc(const Root: string; Operation: TCdDoorOp): Boolean;
{ jen Win NT/2k/XP }
function OpenCloseCdIocNT(const Root: string; Operation: TCdDoorOp): Boolean;
{ jen Win 95/98/Me }
function OpenCloseCdIoc32(const Root: string; Operation: TCdDoorOp): Boolean;

type
  TCdLockOp = (cdUnLock, cdLock);

{ zabrani otevreni dvirek jednotky }
function LockCdIoc(const Root: string; Operation: TCdLockOp): Boolean;
{ jen Win NT/2k/XP }
function LockCdIocNT(const Root: string; Operation: TCdLockOp): Boolean;
{ jen Win 95/98/Me }
function LockCdIoc32(const Root: string; Operation: TCdLockOp): Boolean;

implementation

procedure ListCdDrives(List: TStrings);
var
  Drives: Cardinal;
  i: Integer;
  Root: string;
begin
  Drives := GetLogicalDrives;
  for i := 0 to Ord('Z')-Ord('A')-1 do begin
    if (Drives and 1) = 1 then begin
      Root := Chr(Ord('A')+i)+':\';
      if GetDriveType(PChar(Root)) = DRIVE_CDROM then List.Add(Root);
    end;
    Drives := Drives shr 1;
  end;
end;

function MediumPresent(const Root: string): Boolean;
var
  SavedErrorMode: Cardinal;
begin
  SavedErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
  Result := DirectoryExists(Root);
  SetErrorMode(SavedErrorMode);
end;

procedure GetMediumInfo(const Root: string; var Serial: Cardinal;
  var Name: string);
var
  szName: array [0..MAX_PATH-1] of Char;
  Dummy1, Dummy2: Cardinal;
begin
  if not GetVolumeInformation(PChar(Root), szName, Length(szName), @Serial,
    Dummy1, Dummy2, nil, 0) then
      raise ECdRom.CreateFmt('Chyba Windows: %s', [SysErrorMessage(GetLastError)])
  else
    Name := szName;
end;

procedure OpenCloseCd(const Root: string; Operation: TCdDoorOp);
var
  mciErr: MCIERROR;
  mciParms: TMCI_Open_Parms;
  mciFlags: Cardinal;
begin
  FillChar(mciParms, SizeOf(mciParms), 0);
  with mciParms do begin
    lpstrDeviceType := PChar(MCI_DEVTYPE_CD_AUDIO);
    lpstrElementName := PChar(Root);
  end;
  mciErr := mciSendCommand(0, MCI_OPEN,
    MCI_OPEN_TYPE or MCI_OPEN_TYPE_ID or MCI_OPEN_ELEMENT,
    Cardinal(@mciParms));
  if mciErr = 0 then begin
    if Operation = cdOpen then
      mciFlags := MCI_SET_DOOR_OPEN
    else
      mciFlags := MCI_SET_DOOR_CLOSED;
    mciErr := mciSendCommand(mciParms.wDeviceID, MCI_SET, mciFlags, 0);
    if mciErr <> 0 then raise ECdRom.Create('Chyba pi manipulaci s jednotkou pes MCI.');
    mciErr := mciSendCommand(mciParms.wDeviceID, MCI_CLOSE, 0, Cardinal(@mciParms));
    if mciErr <> 0 then raise ECdRom.Create('Chyba pi zavrn jednotky pes MCI.');
  end
  else
    raise ECdRom.Create('Chyba pi otvrn jednotky pes MCI.');
end;

function DriveLetterFromMask(Mask: Cardinal): Char;
var
  i: Integer;
begin
  i := 0;
  while (i < Ord('Z')-Ord('A')) and ((Mask and 1) = 0) do begin
    Inc(i);
    Mask := Mask shr 1;
  end;
  if i = Ord('Z')-Ord('A') then
    Result := #0
  else
    Result := Chr(Ord('A')+i);
end;

function CdRwPresent(var Root: string): Boolean;
var
  CDBurn: ICDBurn;
  szRoot: array [0..3] of WideChar;
  res: BOOL;
begin
  Root := '';

  CDBurn := CreateComObject(CLSID_CDBurn) as ICDBurn;

  with CDBurn do begin
    OleCheck(HasRecordableDrive(res));
    Result := res;
    if Result then begin
      if Succeeded(GetRecorderDriveLetter(szRoot, Length(szRoot))) then
        Root := szRoot
      else
        Root := '';
    end;
  end;

end;

function CdRwBurn: Boolean;
var
  CDBurn: ICDBurn;
begin
  CDBurn := CreateComObject(CLSID_CDBurn) as ICDBurn;
  Result := Succeeded(CDBurn.Burn(Application.Handle));
end;

function CdRwPath: string;
var
  path: array [0..MAX_PATH-1] of Char;
begin
  if SHGetSpecialFolderPath(Application.Handle, path, CSIDL_CDBURN_AREA, False) then
    Result := path
  else
    Result := '';
end;

function OpenCloseCdIocNT(const Root: string; Operation: TCdDoorOp): Boolean;
var
  hVolume: THandle;
  Returned: Cardinal;
  Removal: BOOL;
  Code: Cardinal;
begin
  hVolume := CreateFile(PChar('\\.\'+Root[1]+':'), GENERIC_READ,
    FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
  if hVolume = INVALID_HANDLE_VALUE then begin
    Result := False;
    Exit;
  end;

  Result := DeviceIoControl(hVolume, FSCTL_LOCK_VOLUME, nil, 0, nil, 0, Returned, nil);
  if Result then begin
    Result := DeviceIoControl(hVolume, FSCTL_DISMOUNT_VOLUME, nil, 0, nil, 0, Returned, nil);
    if Result then begin
      Removal := False;
      Result := DeviceIoControl(hVolume, IOCTL_STORAGE_MEDIA_REMOVAL, @Removal, SizeOf(Removal),
        nil, 0, Returned, nil);
      if Result then begin
        if Operation = cdOpen then
          Code := IOCTL_STORAGE_EJECT_MEDIA
        else
          Code := IOCTL_STORAGE_LOAD_MEDIA;
        Result := DeviceIoControl(hVolume, Code, nil, 0, nil, 0, Returned, nil);
      end;
    end;
  end;

  CloseHandle(hVolume);
end;

function OpenVWin32: THandle;
begin
  Result := CreateFile('\\.\vwin32', 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);
end;

function LockLogicalVolume32(hVWin32: THandle; Drive: Byte): Boolean;
var
  regs: TDIOCRegisters;
  Returned: Cardinal;
begin
  // uzamknout jednotku pro exkluzivni pristup
  with regs do begin
    { kod sluzby (preruseni 21h):
        AH = 44h - I/O control
        AL = 0Dh - generic I/O control }
    EAX := $4A0D;
    {   CH = 08h - kategorie diskove jednotky
        CL = 4Bh - podfunkce lock logical volume }
    ECX := $084A;
    {   BL = cislo logicke jednotky - A: = 1, B: = 2, C: = 3, ...
        BH = lock level = 0 }
    EBX := Drive;
    {   DX = drive permissions = 0 = allow writes }
    EDX := 0;

    EDI := 0;
    ESI := 0;
    Flags := 0;
  end;
  Result := DeviceIoControl(hVWin32, VWIN32_DIOC_DOS_IOCTL, @regs, SizeOf(regs),
    @regs, SizeOf(regs), Returned, nil);
  { CF cleared if successful }
  Result := Result and ((regs.Flags and FLAG_CARRY) = 0);
end;

function UnLockLogicalVolume32(hVWin32: THandle; Drive: Byte): Boolean;
var
  regs: TDIOCRegisters;
  Returned: Cardinal;
begin
  // uzamknout jednotku pro exkluzivni pristup
  with regs do begin
    { kod sluzby (preruseni 21h):
        AH = 44h - I/O control
        AL = 0Dh - generic I/O control }
    EAX := $4A0D;
    {   CH = 08h - kategorie diskove jednotky
        CL = 4Bh - podfunkce unlock logical volume }
    ECX := $086A;
    {   BL = cislo logicke jednotky - A: = 1, B: = 2, C: = 3, ...
        BH = lock level = 0 }
    EBX := Drive;

    EDX := 0;
    EDI := 0;
    ESI := 0;
    Flags := 0;
  end;
  Result := DeviceIoControl(hVWin32, VWIN32_DIOC_DOS_IOCTL, @regs, SizeOf(regs),
    @regs, SizeOf(regs), Returned, nil);
  { CF cleared if successful }
  Result := Result and ((regs.Flags and FLAG_CARRY) = 0);
end;

function UnLockMedia32(hVWin32: THandle; Drive: Byte): Boolean;
var
  regs: TDIOCRegisters;
  Returned: Cardinal;
  param: TParamBlock;
  i: Integer;
begin
  // zjistime kolikrat je CD zamceno
  with param do begin
    bOperation := 2;  // get locking status
    bNumLocks := 0;
  end;

  with regs do begin
    { kod sluzby (preruseni 21h):
        AH = 44h - I/O control
        AL = 0Dh - generic I/O control }
    EAX := $4A0D;
    {   CH = 08h - kategorie diskove jednotky
        CL = 48h - podfunkce set media lock state }
    ECX := $0848;
    {   BL = cislo logicke jednotky - A: = 1, B: = 2, C: = 3, ... }
    EBX := Drive;
    {   DX = ukazatel na blok parametru sluzby }
    EDX := Cardinal(@param);

    EDI := 0;
    ESI := 0;
    Flags := 0;
  end;

  Result := DeviceIoControl(hVWin32, VWIN32_DIOC_DOS_IOCTL, @regs, SizeOf(regs),
    @regs, SizeOf(regs), Returned, nil);

  // operace uspela, pokud neni nastaven carry nebo pokud je nastaven a AX je B0h nebo 1
  if Result then begin
    if (regs.Flags and FLAG_CARRY) <> 0 then begin
      Result := (regs.EAX = $B0) or (regs.EAX = $01);
    end;
  end;

  if not Result then Exit;

  // odemkneme disk tolikat, kolikrat byl zamcen
  for i := 1 to param.bNumLocks do begin
    param.bOperation := 1;  // unlock the media
    with regs do begin
      EAX := $4A0D;
      ECX := $0848;
      EBX := Drive;
      EDX := Cardinal(@param);
      EDI := 0;
      ESI := 0;
      Flags := 0;
    end;
    Result := DeviceIoControl(hVWin32, VWIN32_DIOC_DOS_IOCTL, @regs, SizeOf(regs),
      @regs, SizeOf(regs), Returned, nil);
    if not Result then Break;
  end;
end;

function LockMedia32(hVWin32: THandle; Drive: Byte): Boolean;
var
  regs: TDIOCRegisters;
  Returned: Cardinal;
  param: TParamBlock;
begin
  // zjistime kolikrat je CD zamceno
  with param do begin
    bOperation := 1;  // locking media in drive
    bNumLocks := 0;
  end;

  with regs do begin
    { kod sluzby (preruseni 21h):
        AH = 44h - I/O control
        AL = 0Dh - generic I/O control }
    EAX := $4A0D;
    {   CH = 08h - kategorie diskove jednotky
        CL = 48h - podfunkce set media lock state }
    ECX := $0848;
    {   BL = cislo logicke jednotky - A: = 1, B: = 2, C: = 3, ... }
    EBX := Drive;
    {   DX = ukazatel na blok parametru sluzby }
    EDX := Cardinal(@param);

    EDI := 0;
    ESI := 0;
    Flags := 0;
  end;

  Result := DeviceIoControl(hVWin32, VWIN32_DIOC_DOS_IOCTL, @regs, SizeOf(regs),
    @regs, SizeOf(regs), Returned, nil);
  Result := Result and ((regs.Flags and FLAG_CARRY) = 0);
end;

function EjectMedia32(hVWin32: THandle; Drive: Byte): Boolean;
var
  regs: TDIOCRegisters;
  Returned: Cardinal;
begin
  // vysunout
  with regs do begin
    { kod sluzby (preruseni 21h):
        AH = 44h - I/O control
        AL = 0Dh - generic I/O control }
    EAX := $4A0D;
    {   CH = 08h - kategorie diskove jednotky
        CL = 49h - podfunkce eject media in drive }
    ECX := $0849;
    {   BL = cislo logicke jednotky - A: = 1, B: = 2, C: = 3, ... }
    EBX := Drive;

    EDX := 0;
    EDI := 0;
    ESI := 0;
    Flags := 0;
  end;
  Result := DeviceIoControl(hVWin32, VWIN32_DIOC_DOS_IOCTL, @regs, SizeOf(regs),
    @regs, SizeOf(regs), Returned, nil);
  { CF cleared if successful }
  Result := Result and ((regs.Flags and FLAG_CARRY) = 0);
end;

function LoadMedia32(hVWin32: THandle; Drive: Byte): Boolean;
begin
  Result := True;
  OpenCloseCd(Chr(Drive+64)+':\', cdClose);
end;

function OpenCloseCdIoc32(const Root: string; Operation: TCdDoorOp): Boolean;
var
  hVWin32: THandle;
  Drive: Byte;
begin
  // otevrit vwin32
  hVWin32 := OpenVWin32;
  Result := hVWin32 <> INVALID_HANDLE_VALUE;
  if not Result then Exit;

  Drive := Ord(UpCase(Root[1]))-64;

  Result := LockLogicalVolume32(hVWin32, Drive);
  if not Result then begin
    CloseHandle(hVWin32);
    Exit;
  end;

  Result := UnLockMedia32(hVWin32, Drive);
  if not Result then begin
    UnLockLogicalVolume32(hVWin32, Drive);
    CloseHandle(hVWin32);
    Exit;
  end;

  if Operation = cdOpen then
    Result := EjectMedia32(hVWin32, Drive)
  else
    Result := LoadMedia32(hVWin32, Drive);

  UnLockLogicalVolume32(hVWin32, Drive);
  CloseHandle(hVWin32);

  if not Result then
    SetLastError(ERROR_ACCESS_DENIED);
end;

function OpenCloseCdIoc(const Root: string; Operation: TCdDoorOp): Boolean;
begin
  if GetVersion < $80000000 then
    Result := OpenCloseCdIocNT(Root, Operation)
  else
    Result := OpenCloseCdIoc32(Root, Operation);
end;

function LockCdIocNT(const Root: string; Operation: TCdLockOp): Boolean;
var
  hVolume: THandle;
  Returned: Cardinal;
  Removal: BOOL;
begin
  hVolume := CreateFile(PChar('\\.\'+Root[1]+':'), GENERIC_READ,
    FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
  if hVolume = INVALID_HANDLE_VALUE then begin
    Result := False;
    Exit;
  end;

  Removal := BOOL(Operation);
  Result := DeviceIoControl(hVolume, IOCTL_STORAGE_MEDIA_REMOVAL, @Removal,
    SizeOf(Removal), nil, 0, Returned, nil);
  CloseHandle(hVolume);
end;

function LockCdIoc32(const Root: string; Operation: TCdLockOp): Boolean;
{var
  hVWin32: THandle;
  Drive: Byte;}
begin
  {hVWin32 := OpenVWin32;
  if hVWin32 = INVALID_HANDLE_VALUE then begin
    Result := False;
    Exit;
  end;

  Drive := Ord(UpCase(Root[1]))-64;

  if Operation = cdLock then begin
    Result := LockMedia32(hVWin32, Drive);
  end
  else begin
    Result := UnLockMedia32(hVWin32, Drive);
  end;

  CloseHandle(hVWin32);}

  Result := False;
  SetLastError(ERROR_NOT_SUPPORTED);
end;

function LockCdIoc(const Root: string; Operation: TCdLockOp): Boolean;
begin
  if GetVersion < $80000000 then
    Result := LockCdIocNT(Root, Operation)
  else
    Result := LockCdIoc32(Root, Operation);
end;

end.
