unit main;

{-------------------------------------------------------------------------------
  main.pas
  hlavni unit demo projektu 'CD-ROM, CD-ROM...'
  2003 (c) manison
-------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TfmCdRom = class(TForm)
    btList: TButton;
    cbList: TComboBox;
    btCdPresent: TButton;
    btOpenMCI: TButton;
    btCloseMCI: TButton;
    chCancelAutoplay: TCheckBox;
    Label1: TLabel;
    memoStatus: TMemo;
    btHasCDBurn: TButton;
    btBurn: TButton;
    btStaging: TButton;
    btOpenIoc: TButton;
    btCloseIoc: TButton;
    btLock: TButton;
    btUnLock: TButton;
    procedure btListClick(Sender: TObject);
    procedure btCdPresentClick(Sender: TObject);
    procedure btOpenMCIClick(Sender: TObject);
    procedure btCloseMCIClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btHasCDBurnClick(Sender: TObject);
    procedure btBurnClick(Sender: TObject);
    procedure btStagingClick(Sender: TObject);
    procedure btCloseIocClick(Sender: TObject);
    procedure btOpenIocClick(Sender: TObject);
    procedure btLockClick(Sender: TObject);
    procedure btUnLockClick(Sender: TObject);
  protected
    procedure WndProc(var Msg: TMessage); override;
  private
    msgCancelAutoplay: UINT;
    AlreadyActivated: Boolean;
    function CheckDrive: Boolean;
    function CheckWinXP: Boolean;
    procedure WMDeviceChange(var Msg: TMessage); message WM_DEVICECHANGE;
  end;

var
  fmCdRom: TfmCdRom;

implementation

uses CDTools;

{$R *.DFM}

procedure TfmCdRom.btListClick(Sender: TObject);
begin
  Screen.Cursor := crHourGlass;
  cbList.Items.BeginUpdate;
  cbList.Clear;
  try
    ListCdDrives(cbList.Items);
  finally
    cbList.Items.EndUpdate;
    Screen.Cursor := crDefault;
  end;
  if cbList.Items.Count = 0 then
    ShowMessage('Na tomto potai nejsou nainstalovny dn jednotky CDROM.')
  else
    cbList.ItemIndex := 0;
end;

procedure TfmCdRom.btCdPresentClick(Sender: TObject);
var
  Serial: Cardinal;
  Name: string;
begin
  if not CheckDrive then Exit;

  if MediumPresent(cbList.Items[cbList.ItemIndex]) then begin
     GetMediumInfo(cbList.Items[cbList.ItemIndex], Serial, Name);
     ShowMessage(Format('V jednotce %s je disk %s se sriovm slem %.8X',
       [cbList.Items[cbList.ItemIndex], Name, Serial]));
  end
  else
    ShowMessage(Format('V jednotce %s nen vloen disk.',
      [cbList.Items[cbList.ItemIndex]]));
end;

procedure TfmCdRom.btOpenMCIClick(Sender: TObject);
begin
  if not CheckDrive then Exit;

  OpenCloseCd(cbList.Items[cbList.ItemIndex], cdOpen);
end;

procedure TfmCdRom.btCloseMCIClick(Sender: TObject);
begin
  if not CheckDrive then Exit;

  OpenCloseCd(cbList.Items[cbList.ItemIndex], cdClose);
end;

function TfmCdRom.CheckDrive: Boolean;
begin
  Result := cbList.ItemIndex <> -1;
  if not Result then
    ShowMessage('Nejdv vyber jednotku ze seznamu.');
end;

procedure TfmCdRom.FormActivate(Sender: TObject);
begin
  if not AlreadyActivated then begin
    AlreadyActivated := True;
    btList.Click;
  end;
end;

procedure TfmCdRom.WMDeviceChange(var Msg: TMessage);
var
  PVolume: PDevBroadcastVolume;
  Drive: Char;
begin
  Msg.Result := 1;

  PVolume := PDevBroadcastVolume(Msg.LParam);
  if PVolume^.dbcv_devicetype <> DBT_DEVTYP_VOLUME then Exit;
  if PVolume^.dbcv_flags <> DBTF_MEDIA then Exit;

  Drive := DriveLetterFromMask(PVolume^.dbcv_unitmask);
  if Drive = #0 then Exit;

  case Msg.WParam of
    DBT_DEVICEARRIVAL: begin
      memoStatus.Lines.Add(Format('> Do jednotky %s:\ byl vloen disk.', [Drive]));
    end;

    DBT_DEVICEREMOVECOMPLETE: begin
      memoStatus.Lines.Add(Format('> Z jednotky %s:\ byl vyjmut disk.', [Drive]));
    end;
  end;
end;

procedure TfmCdRom.WndProc(var Msg: TMessage);
const
  Negator: array[Boolean] of string = ('ne', '');
begin
  if Msg.Msg = msgCancelAutoplay then begin
    Msg.Result := Integer(chCancelAutoplay.Checked);
    memoStatus.Lines.Add(Format('> Bylo vloeno CD. Zvolili jste, e '+
      '%sbude zrueno automatick pehrn disku CD.',
      [Negator[chCancelAutoplay.Checked]]));
  end
  else
    inherited;
end;

procedure TfmCdRom.FormCreate(Sender: TObject);
begin
  msgCancelAutoplay := RegisterWindowMessage('QueryCancelAutoPlay');
end;

procedure TfmCdRom.btHasCDBurnClick(Sender: TObject);
var
  Root: string;
begin
  if not CheckWinXP then Exit;

  if CdRwPresent(Root) then
    if Root <> '' then
      ShowMessage(Format('%s je vypalovac jednotka.', [Root]))
    else
      ShowMessage(Format('Na tomto potai je nainstalovna jedna nebo vce '+
        'jednotek pro zpis disk CD,'#13'dn vak nen aktivovna.', [Root]))
  else
    ShowMessage('Na tomto potai nejsou instalovny dn jednotky pro zpis disk CD.');
end;

function TfmCdRom.CheckWinXP: Boolean;
var
  ver: TOSVersionInfo;
begin
  ver.dwOSVersionInfoSize := SizeOf(ver);
  GetVersionEx(ver);
  Result := (ver.dwMajorVersion > 5) or ((ver.dwMajorVersion = 5) and
    (ver.dwMinorVersion >= 1));
  if not Result then
    ShowMessage('Tuto funkci mete pout ve verzi Windows XP a vy.');
end;

procedure TfmCdRom.btBurnClick(Sender: TObject);
begin
  if not CheckWinXP then Exit;

  if not CdRwBurn then
    ShowMessage('Zpis soubor na disk CD se nezdail.');
end;

procedure TfmCdRom.btStagingClick(Sender: TObject);
var
  CDBurnPath: string;
begin
  if not CheckWinXP then Exit;

  CDBurnPath := CdRwPath;

  ShowMessage('Sloka pro uloen soubor ped vyplenm na CD:'#13+CDBurnPath);
end;

procedure TfmCdRom.btCloseIocClick(Sender: TObject);
begin
  if not CheckDrive then Exit;

  if not OpenCloseCdIoc(cbList.Items[cbList.ItemIndex], cdClose) then
    ShowMessage('Dvka nelze zavt protoe '+AnsiLowerCase(SysErrorMessage(GetLastError)));
end;

procedure TfmCdRom.btOpenIocClick(Sender: TObject);
begin
  if not CheckDrive then Exit;

  if not OpenCloseCdIoc(cbList.Items[cbList.ItemIndex], cdOpen) then
    ShowMessage('Dvka nelze otevt protoe '+AnsiLowerCase(SysErrorMessage(GetLastError)));
end;

procedure TfmCdRom.btLockClick(Sender: TObject);
begin
  if not CheckDrive then Exit;

  if not LockCdIoc(cbList.Items[cbList.ItemIndex], cdLock) then
    ShowMessage('Dvka nelze zamknout protoe '+AnsiLowerCase(SysErrorMessage(GetLastError)));
end;

procedure TfmCdRom.btUnLockClick(Sender: TObject);
begin
  if not CheckDrive then Exit;

  if not LockCdIoc(cbList.Items[cbList.ItemIndex], cdUnLock) then
    ShowMessage('Dvka nelze zamknout protoe '+AnsiLowerCase(SysErrorMessage(GetLastError)));
end;

end.
